from argparse import Namespace

import torch
import torch.nn as nn
import re

class AttrProxy(object):
    """Translates index lookups into attribute lookups"""
    def __init__(self, module, prefix):
        self.module = module
        self.prefix = prefix

    def __len__(self):
        return len([x for x in self.module.__dict__['_modules'].keys() if re.match(f'{self.prefix}\d+', x)])

    def __getitem__(self, item):
        if item >= len(self):
            raise IndexError
        return getattr(self.module, self.prefix + str(item))

class MultiReadout(nn.Module):
    """A fake list of FFNs for reading out as suggested in
    https://discuss.pytorch.org/t/list-of-nn-module-in-a-nn-module/219/3 """

    def __init__(self, args: Namespace, atom_targets, bond_targets=None,
                 atom_constraints=None, bond_constraints=None, attention=False):
        """

        :param args:
        :param args:
        :param constraints:
        """

        features_size = args.hidden_size
        hidden_size = args.ffn_hidden_size
        num_layers = args.ffn_num_layers
        output_size = args.output_size
        dropout = nn.Dropout(args.dropout)
        activation = nn.ReLU()

        super(MultiReadout, self).__init__()
        for i, a_target in enumerate(atom_targets):
            constraint = atom_constraints[i] if atom_constraints is not None and i < len(atom_constraints) else None
            if attention:
                self.add_module(f'readout_{i}', FFNAtten(features_size, hidden_size, num_layers,
                                                         output_size, dropout, activation, constraint, ffn_type='atom'))
            else:
                self.add_module(f'readout_{i}', FFN(features_size, hidden_size, num_layers,
                                                    output_size, dropout, activation, constraint, ffn_type='atom'))

        i += 1
        for j, b_target in enumerate(bond_targets):
            i += j
            constraint = bond_constraints[i] if bond_constraints and j < len(bond_constraints) else None
            self.add_module(f'readout_{i}', FFN(features_size, hidden_size, num_layers,
                                                output_size, dropout, activation, constraint, ffn_type='bond'))

        self.ffn_list = AttrProxy(self, 'readout_')

    def forward(self, *input):
        return [ffn(*input) for ffn in self.ffn_list]


class FFNAtten(nn.Module):
    def __init__(self, features_size, hidden_size, num_layers, output_size,
                 dropout, activation, constraint=None, ffn_type='atom'):
        """Initializes the FFN.

        args: Arguments.
        constraints: constraints applied to output
        """

        super(FFN, self).__init__()

        self.ffn = DenseLayers(features_size, hidden_size,
                               num_layers, hidden_size, dropout, activation)
        self.ffn_readout = DenseLayers(hidden_size, hidden_size, 1, output_size, dropout, activation)


        self.weights_readout = DenseLayers(first_linear_dim=hidden_size, output_size=output_size, num_layers=2,
                                           dropout=dropout, activation=activation)
        self.constraint = constraint

    def forward(self, input):
        """
        Runs the FFN on input

        :param input:
        :return:
        """
        a_hidden, a_scope, b_hidden, b_scope, b2br, bond_types = input

        hidden = a_hidden
        scope = a_scope

        output_hidden = self.ffn(hidden)
        output = self.ffn_readout(output_hidden)

        weights = self.weights_readout(output_hidden)
        constrained_output = []
        for i, (a_start, a_size) in enumerate(scope):
            if a_size == 0:
                continue
            else:
                cur_weights = weights.narrow(0, a_start, a_size)
                cur_output = output.narrow(0, a_start, a_size)

                cur_weights = torch.nn.Softmax()(cur_weights)

                cur_output_sum = cur_output.sum()

                cur_output = cur_output + cur_weights * (self.constraint - cur_output_sum)
                constrained_output.append(cur_output)
        output = torch.cat(constrained_output, dim=0)

        return output


class FFN(nn.Module):
    """A Feedforward netowrk reading out properties from fingerprint"""

    def __init__(self, features_size, hidden_size, num_layers, output_size,
                 dropout, activation, constraint=None, ffn_type='atom', attention=False):
        """Initializes the FFN.

        args: Arguments.
        constraints: constraints applied to output
        """

        super(FFN, self).__init__()
        if ffn_type == 'atom':
            self.ffn = DenseLayers(features_size, hidden_size,
                                   num_layers, output_size, dropout, activation)
        elif ffn_type == 'bond':
            self.ffn = DenseLayers(2*features_size, hidden_size,
                                   num_layers, output_size, dropout, activation)
        self.ffn_type = ffn_type
        self.attention = attention

        if constraint is not None:
            self.weights_readout = DenseLayers(features_size, hidden_size,
                                               num_layers, output_size, dropout, activation)
            if attention:
                self.weights_readout = DenseLayers(first_linear_dim=hidden_size, output_size=1, num_layers=1,
                                                   dropout=dropout, activation=activation)
            self.constraint = constraint
        else:
            self.constraint = None

    def forward(self, input):
        """
        Runs the FFN on input

        :param input:
        :return:
        """
        a_hidden, a_scope, b_hidden, b_scope, b2br, bond_types = input

        if self.ffn_type == 'atom':
            hidden = a_hidden
            scope = a_scope

            output = self.ffn(hidden)
            if self.attention:
                weights = self.weights_readout(output)


            if self.constraint is not None:
                weights = self.weights_readout(hidden)
                constrained_output = []
                for i, (a_start, a_size) in enumerate(scope):
                    if a_size == 0:
                        continue
                    else:
                        cur_weights = weights.narrow(0, a_start, a_size)
                        cur_output = output.narrow(0, a_start, a_size)

                        cur_weights_sum = cur_weights.sum()
                        cur_output_sum = cur_output.sum()

                        cur_output = cur_output + cur_weights * \
                                     (self.constraint - cur_output_sum) / cur_weights_sum
                        constrained_output.append(cur_output)
                output = torch.cat(constrained_output, dim=0)
            else:
                output = output[1:]
        elif self.ffn_type == 'bond':

            forward_bond = b_hidden[b2br[:, 0]]
            backward_bond = b_hidden[b2br[:, 1]]

            b_hidden = torch.cat([forward_bond, backward_bond], dim=1)

            output = self.ffn(b_hidden) + bond_types.reshape(-1, 1)

        return output


class DenseLayers(nn.Module):
    "Dense layers"

    def __init__(self,
                 first_linear_dim: int,
                 hidden_size: int,
                 num_layers: int,
                 output_size: int,
                 dropout: nn.Module,
                 activation) -> nn.Sequential:
        """
        :param first_linear_dim:
        :param hidden_size:
        :param num_layers:
        :param output_size:
        :param dropout:
        :param activation:
        """
        super(DenseLayers, self).__init__()
        if num_layers == 1:
            layers = [
                dropout,
                nn.Linear(first_linear_dim, output_size)
            ]
        else:
            layers = [
                dropout,
                nn.Linear(first_linear_dim, hidden_size)
            ]
            for _ in range(num_layers - 2):
                layers.extend([
                    activation,
                    dropout,
                    nn.Linear(hidden_size, hidden_size),
                ])
            layers.extend([
                activation,
                dropout,
                nn.Linear(hidden_size, output_size),
            ])

        self.dense_layers = nn.Sequential(*layers)

    def forward(self, input):
        return self.dense_layers(input)



import os

os.environ['KMP_DUPLICATE_LIB_OK']='True'

from argparse import Namespace

import torch
import torch.nn as nn

from qmdesc.nn_utils import index_select_ND


class MPNEncoder(nn.Module):
    """A message passing neural network for encoding a molecule."""

    def __init__(self, args: Namespace, atom_fdim: int, bond_fdim: int):
        """Initializes the MPNEncoder.

        :param args: Arguments.
        :param atom_fdim: Atom features dimension.
        :param bond_fdim: Bond features dimension.
        """
        super(MPNEncoder, self).__init__()
        self.atom_fdim = atom_fdim
        self.bond_fdim = bond_fdim
        self.hidden_size = args.hidden_size
        self.bias = args.bias
        self.depth = args.depth
        self.dropout = args.dropout
        self.layers_per_message = 1
        self.undirected = args.undirected
        self.atom_messages = args.atom_messages
        self.features_only = args.features_only
        self.use_input_features = args.use_input_features
        self.args = args

        if self.features_only:
            return

        # Dropout
        self.dropout_layer = nn.Dropout(p=self.dropout)

        # Activation
        self.act_func = nn.ReLU()

        # Cached zeros
        self.cached_zero_vector = nn.Parameter(torch.zeros(self.hidden_size), requires_grad=False)

        # Input
        input_dim = self.atom_fdim if self.atom_messages else self.bond_fdim
        self.W_i = nn.Linear(input_dim, self.hidden_size, bias=self.bias)

        if self.atom_messages:
            w_h_input_size = self.hidden_size + self.bond_fdim
        else:
            w_h_input_size = self.hidden_size

        # Shared weight matrix across depths (default)
        self.W_h = nn.Linear(w_h_input_size, self.hidden_size, bias=self.bias)

        # hidden state readout
        self.W_o_a = nn.Linear(self.atom_fdim + self.hidden_size, self.hidden_size)
        self.W_o_b = nn.Linear(self.bond_fdim + self.hidden_size, self.hidden_size)

    def forward(self,
                inputs) -> torch.FloatTensor:
        """
        Encodes a batch of molecular graphs.
        :return: A PyTorch tensor of shape (num_molecules, hidden_size) containing the encoding of each molecule.
        """

        f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, b2br, bond_types = inputs

        input = self.W_i(f_bonds)  # num_bonds x hidden_size
        message = self.act_func(input)  # num_bonds x hidden_size

        # Message passing
        for depth in range(self.depth - 1):
            if self.undirected:
                message = (message + message[b2revb]) / 2


            # m(a1 -> a2) = [sum_{a0 \in nei(a1)} m(a0 -> a1)] - m(a2 -> a1)
            # message      a_message = sum(nei_a_message)      rev_message
            nei_a_message = index_select_ND(message, a2b)  # num_atoms x max_num_bonds x hidden
            a_message = nei_a_message.sum(dim=1)  # num_atoms x hidden
            rev_message = message[b2revb]  # num_bonds x hidden
            message = a_message[b2a] - rev_message  # num_bonds x hidden

            message = self.W_h(message)
            message = self.act_func(input + message)  # num_bonds x hidden_size
            message = self.dropout_layer(message)  # num_bonds x hidden

        # atom hidden
        a2x = a2b
        nei_a_message = index_select_ND(message, a2x)  # num_atoms x max_num_bonds x hidden
        a_message = nei_a_message.sum(dim=1)  # num_atoms x hidden
        a_input = torch.cat([f_atoms, a_message], dim=1)  # num_atoms x (atom_fdim + hidden)
        atom_hiddens = self.act_func(self.W_o_a(a_input))  # num_atoms x hidden
        atom_hiddens = self.dropout_layer(atom_hiddens)  # num_atoms x hidden

        # bond hidden
        b_input = torch.cat([f_bonds, message], dim=1)
        bond_hiddens = self.act_func(self.W_o_b(b_input))
        bond_hiddens = self.dropout_layer(bond_hiddens)

        return atom_hiddens, a_scope, bond_hiddens, b_scope, b2br, bond_types  # num_atoms x hidden, remove the first one which is zero padding


class MPN(nn.Module):
    def __init__(self, args, atom_fdim, bond_fdim):
        super(MPN, self).__init__()
        self.encoder = MPNEncoder(args, atom_fdim, bond_fdim)

    def forward(self, *inputs):
        return self.encoder(*inputs)



"""
This module defines the PathwayRankingHandler for use in Torchserve.
"""
from typing import Dict

import pkg_resources
import torch
from qmdesc.featurization import mol2graph, get_atom_fdim, get_bond_fdim
from rdkit import Chem

class ReactivityDescriptorHandler():
    '''Wrap the trained atom-bond qm descriptors predicting model

    Predict QM descriptors for a given SMILES string of organic compound containing C, H, O, N, P, S, F, Cl, Br, I, B

    Example:
            >>> from qmdesc import ReactivityDescriptorHandler
            >>> handler = ReactivityDescriptorHandler()
            >>> results = handler.predict('CCCC')
    '''

    def __init__(self):
        """
        ReactivityDescriptorHandler constructor.
        """
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        model_pt_path = "QM_137k.pt"

        from qmdesc.model import MoleculeModel

        # Load model and args
        stream = pkg_resources.resource_stream(__name__, model_pt_path)
        state = torch.load(stream, lambda storage, loc: storage)
        args, loaded_state_dict = state['args'], state['state_dict']
        atom_fdim = get_atom_fdim()
        bond_fdim = get_bond_fdim() + atom_fdim

        self.model = MoleculeModel(args, atom_fdim, bond_fdim)
        self.model.load_state_dict(loaded_state_dict)
        self.model.to(self.device)
        self.model.eval()

        self.initalized = True

    def _preprocess(self, smiles: str):
        """
        Preprocess SMILES

        :param smiles: SMILES string
        :return: molecular graph
        """
        mol_graph = mol2graph(smiles)
        f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, b2br, bond_types = mol_graph.get_components()
        f_atoms, f_bonds, a2b, b2a, b2revb, b2br, bond_types = \
            f_atoms.to(self.device), f_bonds.to(self.device), a2b.to(self.device), b2a.to(self.device), \
            b2revb.to(self.device), b2br.to(self.device), bond_types.to(self.device)

        return f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, b2br, bond_types

    def _inference(self, data):
        """
        model prediction

        :param data: molecular graph
        :return: The output of the model
        """
        descs = self.model(data)

        return descs

    def _postprocess(self, inference_output) -> Dict:
        """
        Postprocess results

        :param inference_output: The output of the model
        :return: Results
        """

        smiles = inference_output['smiles']
        descs = inference_output['descs']

        descs = [x.data.cpu().numpy() for x in descs]

        partial_charge, partial_neu, partial_elec, NMR, bond_order, bond_distance = descs

        results = {'smiles': smiles, 'partial_charge': partial_charge.flatten(), 'fukui_neu': partial_neu.flatten(),
                   'fukui_elec': partial_elec.flatten(), 'NMR': NMR.flatten(), 'bond_order': bond_order.flatten(),
                   'bond_length': bond_distance.flatten()}
        return results

    def predict(self,
                smiles: str,
                sdf: str = None) -> Dict:
        """
        Wrap the preprocess, inference, and postprocess

        :param smiles: Input SMILES string
        :param sdf: Output .sdf file
        :return: A dictionary containing the prediction result
        """

        outputs = self._inference(self._preprocess([smiles]))
        postprocess_inputs = {'smiles': smiles, 'descs': outputs}
        results = self._postprocess(postprocess_inputs)

        if sdf is not None:
            if not sdf.endswith('.sdf'):
                print('must provide a sdf name end up with \'.sdf\'')
                return results

            writer = Chem.SDWriter(sdf)
            m = Chem.MolFromSmiles(smiles)
            m = Chem.AddHs(m)

            for p in results:
                p_upper = p.upper()
                if p == 'smiles':
                    m.SetProp(p_upper, results[p])
                else:
                    m.SetProp(p_upper, ','.join(str(x) for x in results[p]))

            name = sdf.strip('.sdf')
            m.SetProp('_Name', name)

            writer.write(m)

        return results

def qmdesc() -> None:
    """
    This is the entry point for the command line command :code:'qmdesc'

    Example:
       $ qmdesc CCCC --sdf CCCC.sdf
    """
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('smiles', type=str,
                        help='Input smiles string')
    parser.add_argument('--sdf', default='qmdesc.sdf', type=str,
                        help='output sdf saving the qm descriptors')
    args = parser.parse_args()

    predictor = ReactivityDescriptorHandler()
    results = predictor.predict(args.smiles, sdf=args.sdf)





from typing import List, Tuple, Union

from rdkit import Chem
import torch
import numpy as np

# Atom feature sizes
MAX_ATOMIC_NUM = 100
ATOM_FEATURES = {
    'atomic_num': list(range(MAX_ATOMIC_NUM)),
    'degree': [0, 1, 2, 3, 4, 5],
    'formal_charge': [-1, -2, 1, 2, 0],
    'chiral_tag': [0, 1, 2, 3],
    'num_Hs': [0, 1, 2, 3, 4],
    'hybridization': [
        Chem.rdchem.HybridizationType.SP,
        Chem.rdchem.HybridizationType.SP2,
        Chem.rdchem.HybridizationType.SP3,
        Chem.rdchem.HybridizationType.SP3D,
        Chem.rdchem.HybridizationType.SP3D2
    ],
}

# Distance feature sizes
PATH_DISTANCE_BINS = list(range(10))
THREE_D_DISTANCE_MAX = 20
THREE_D_DISTANCE_STEP = 1
THREE_D_DISTANCE_BINS = list(range(0, THREE_D_DISTANCE_MAX + 1, THREE_D_DISTANCE_STEP))

# len(choices) + 1 to include room for uncommon values; + 2 at end for IsAromatic and mass
ATOM_FDIM = sum(len(choices) + 1 for choices in ATOM_FEATURES.values()) + 2
BOND_FDIM = 14

# Memoization
SMILES_TO_GRAPH = {}


def clear_cache():
    """Clears featurization cache."""
    global SMILES_TO_GRAPH
    SMILES_TO_GRAPH = {}


def get_atom_fdim() -> int:
    """
    Gets the dimensionality of atom features.
    :param: Arguments.
    """
    return ATOM_FDIM


def get_bond_fdim() -> int:
    """
    Gets the dimensionality of bond features.
    :param: Arguments.
    """
    return BOND_FDIM


def onek_encoding_unk(value: int, choices: List[int]) -> List[int]:
    """
    Creates a one-hot encoding.
    :param value: The value for which the encoding should be one.
    :param choices: A list of possible values.
    :return: A one-hot encoding of the value in a list of length len(choices) + 1.
    If value is not in the list of choices, then the final element in the encoding is 1.
    """
    encoding = [0] * (len(choices) + 1)
    index = choices.index(value) if value in choices else -1
    encoding[index] = 1

    return encoding


def atom_features(atom: Chem.rdchem.Atom, functional_groups: List[int] = None) -> List[Union[bool, int, float]]:
    """
    Builds a feature vector for an atom.
    :param atom: An RDKit atom.
    :param functional_groups: A k-hot vector indicating the functional groups the atom belongs to.
    :return: A list containing the atom features.
    """
    features = onek_encoding_unk(atom.GetAtomicNum() - 1, ATOM_FEATURES['atomic_num']) + \
               onek_encoding_unk(atom.GetTotalDegree(), ATOM_FEATURES['degree']) + \
               onek_encoding_unk(atom.GetFormalCharge(), ATOM_FEATURES['formal_charge']) + \
               onek_encoding_unk(int(atom.GetChiralTag()), ATOM_FEATURES['chiral_tag']) + \
               onek_encoding_unk(int(atom.GetTotalNumHs()), ATOM_FEATURES['num_Hs']) + \
               onek_encoding_unk(int(atom.GetHybridization()), ATOM_FEATURES['hybridization']) + \
               [1 if atom.GetIsAromatic() else 0] + \
               [atom.GetMass() * 0.01]  # scaled to about the same range as other features
    if functional_groups is not None:
        features += functional_groups
    return features


def bond_features(bond: Chem.rdchem.Bond) -> List[Union[bool, int, float]]:
    """
    Builds a feature vector for a bond.
    :param bond: A RDKit bond.
    :return: A list containing the bond features.
    """
    if bond is None:
        fbond = [1] + [0] * (BOND_FDIM - 1)
    else:
        bt = bond.GetBondType()
        fbond = [
            0,  # bond is not None
            bt == Chem.rdchem.BondType.SINGLE,
            bt == Chem.rdchem.BondType.DOUBLE,
            bt == Chem.rdchem.BondType.TRIPLE,
            bt == Chem.rdchem.BondType.AROMATIC,
            (bond.GetIsConjugated() if bt is not None else 0),
            (bond.IsInRing() if bt is not None else 0)
        ]
        fbond += onek_encoding_unk(int(bond.GetStereo()), list(range(6)))
    return fbond


class MolGraph:
    """
    A MolGraph represents the graph structure and featurization of a single molecule.
    A MolGraph computes the following attributes:
    - smiles: Smiles string.
    - n_atoms: The number of atoms in the molecule.
    - n_bonds: The number of bonds in the molecule.
    - f_atoms: A mapping from an atom index to a list atom features.
    - f_bonds: A mapping from a bond index to a list of bond features.
    - a2b: A mapping from an atom index to a list of incoming bond indices.
    - b2a: A mapping from a bond index to the index of the atom the bond originates from.
    - b2revb: A mapping from a bond index to the index of the reverse bond.
    """

    def __init__(self, smiles: str):
        """
        Computes the graph structure and featurization of a molecule.
        :param smiles: A smiles string.
        """
        self.smiles = smiles
        self.n_atoms = 0  # number of atoms
        self.n_bonds = 0  # number of bonds
        self.f_atoms = []  # mapping from atom index to atom features
        self.f_bonds = []  # mapping from bond index to concat(in_atom, bond) features
        self.a2b = []  # mapping from atom index to incoming bond indices
        self.b2a = []  # mapping from bond index to the index of the atom the bond is coming from
        self.b2revb = []
        # mapping from bond index to the index of the reverse bond

        # Convert smiles to molecule
        mol = Chem.MolFromSmiles(smiles)

        # Add explicit Hs
        mol = Chem.AddHs(mol)

        # fake the number of "atoms" if we are collapsing substructures
        self.n_atoms = mol.GetNumAtoms()

        # mapping from real bond index (the same as in target) to the f_bonds, since each real has two bond features,
        # only save the index of first one, the second one should be the first_index + 1
        self.b2br = np.zeros([len(mol.GetBonds()), 2])
        self.bond_types = [b.GetBondTypeAsDouble() for b in mol.GetBonds()]

        # Get atom features
        for i, atom in enumerate(mol.GetAtoms()):
            self.f_atoms.append(atom_features(atom))
        self.f_atoms = [self.f_atoms[i] for i in range(self.n_atoms)]

        for _ in range(self.n_atoms):
            self.a2b.append([])

        # Get bond features
        for a1 in range(self.n_atoms):
            for a2 in range(a1 + 1, self.n_atoms):
                bond = mol.GetBondBetweenAtoms(a1, a2)

                if bond is None:
                    continue

                f_bond = bond_features(bond)

                self.f_bonds.append(self.f_atoms[a1] + f_bond)
                self.f_bonds.append(self.f_atoms[a2] + f_bond)

                # Update index mappings
                b1 = self.n_bonds
                b2 = b1 + 1
                self.a2b[a2].append(b1)  # b1 = a1 --> a2
                self.b2a.append(a1)
                self.a2b[a1].append(b2)  # b2 = a2 --> a1
                self.b2a.append(a2)
                self.b2revb.append(b2)
                self.b2revb.append(b1)
                self.b2br[bond.GetIdx(), :] = [self.n_bonds, self.n_bonds + 1]
                self.n_bonds += 2


class BatchMolGraph:
    """
    A BatchMolGraph represents the graph structure and featurization of a batch of molecules.
    A BatchMolGraph contains the attributes of a MolGraph plus:
    - smiles_batch: A list of smiles strings.
    - n_mols: The number of molecules in the batch.
    - atom_fdim: The dimensionality of the atom features.
    - bond_fdim: The dimensionality of the bond features (technically the combined atom/bond features).
    - a_scope: A list of tuples indicating the start and end atom indices for each molecule.
    - b_scope: A list of tuples indicating the start and end bond indices for each molecule.
    - max_num_bonds: The maximum number of bonds neighboring an atom in this batch.
    - b2b: (Optional) A mapping from a bond index to incoming bond indices.
    - a2a: (Optional): A mapping from an atom index to neighboring atom indices.
    """

    def __init__(self, mol_graphs: List[MolGraph]):
        self.smiles_batch = [mol_graph.smiles for mol_graph in mol_graphs]
        self.n_mols = len(self.smiles_batch)

        self.atom_fdim = get_atom_fdim()
        self.bond_fdim = get_bond_fdim() + self.atom_fdim

        # Start n_atoms and n_bonds at 1 b/c zero padding
        self.n_atoms = 1  # number of atoms (start at 1 b/c need index 0 as padding)
        self.n_bonds = 1  # number of bonds (start at 1 b/c need index 0 as padding)
        self.a_scope = []  # list of tuples indicating (start_atom_index, num_atoms) for each molecule
        self.b_scope = []  # list of tuples indicating (start_bond_index, num_bonds) for each molecule

        # All start with zero padding so that indexing with zero padding returns zeros
        f_atoms = [[0] * self.atom_fdim]  # atom features
        f_bonds = [[0] * self.bond_fdim]  # combined atom/bond features
        a2b = [[]]  # mapping from atom index to incoming bond indices
        b2a = [0]  # mapping from bond index to the index of the atom the bond is coming from
        b2revb = [0]  # mapping from bond index to the index of the reverse bond
        b2br = []  # mapping from f_bdons to real bonds in molecule recorded in target

        # used as add-on for bond index
        bond_types = []

        # used for final add up of bond hiddens (because no unsorted_segment_sum in pytorch)
        for mol_graph in mol_graphs:
            f_atoms.extend(mol_graph.f_atoms)
            f_bonds.extend(mol_graph.f_bonds)

            for a in range(mol_graph.n_atoms):
                a2b.append([b + self.n_bonds for b in mol_graph.a2b[a]])

            for b in range(mol_graph.n_bonds):
                b2a.append(self.n_atoms + mol_graph.b2a[b])
                b2revb.append(self.n_bonds + mol_graph.b2revb[b])

            self.a_scope.append((self.n_atoms, mol_graph.n_atoms))
            self.b_scope.append((self.n_bonds, mol_graph.n_bonds))
            b2br.append(mol_graph.b2br + self.n_bonds)

            self.n_atoms += mol_graph.n_atoms
            self.n_bonds += mol_graph.n_bonds

            bond_types.extend(mol_graph.bond_types)

        b2br = np.concatenate(b2br, axis=0)

        self.max_num_bonds = max(1, max(
            len(in_bonds) for in_bonds in a2b))  # max with 1 to fix a crash in rare case of all single-heavy-atom mols

        self.f_atoms = torch.FloatTensor(f_atoms)
        self.f_bonds = torch.FloatTensor(f_bonds)
        self.a2b = torch.LongTensor([a2b[a] + [0] * (self.max_num_bonds - len(a2b[a])) for a in range(self.n_atoms)])
        self.b2a = torch.LongTensor(b2a)
        self.b2revb = torch.LongTensor(b2revb)
        self.b2br = torch.LongTensor(b2br)
        self.bond_types = torch.FloatTensor(bond_types)
        self.b2b = None  # try to avoid computing b2b b/c O(n_atoms^3)
        self.a2a = None  # only needed if using atom messages

    def get_components(self) -> Tuple[torch.FloatTensor, torch.FloatTensor,
                                      torch.LongTensor, torch.LongTensor, torch.LongTensor,
                                      List[Tuple[int, int]], List[Tuple[int, int]], torch.LongTensor,
                                      torch.FloatTensor]:
        """
        Returns the components of the BatchMolGraph.
        :return: A tuple containing PyTorch tensors with the atom features, bond features, and graph structure
        and two lists indicating the scope of the atoms and bonds (i.e. which molecules they belong to).
        """
        return self.f_atoms, self.f_bonds, self.a2b, self.b2a, self.b2revb, \
               self.a_scope, self.b_scope, self.b2br, self.bond_types

    def get_b2b(self) -> torch.LongTensor:
        """
        Computes (if necessary) and returns a mapping from each bond index to all the incoming bond indices.
        :return: A PyTorch tensor containing the mapping from each bond index to all the incoming bond indices.
        """

        if self.b2b is None:
            b2b = self.a2b[self.b2a]  # num_bonds x max_num_bonds
            # b2b includes reverse edge for each bond so need to mask out
            revmask = (b2b != self.b2revb.unsqueeze(1).repeat(1, b2b.size(1))).long()  # num_bonds x max_num_bonds
            self.b2b = b2b * revmask

        return self.b2b

    def get_a2a(self) -> torch.LongTensor:
        """
        Computes (if necessary) and returns a mapping from each atom index to all neighboring atom indices.
        :return: A PyTorch tensor containing the mapping from each bond index to all the incodming bond indices.
        """
        if self.a2a is None:
            # b = a1 --> a2
            # a2b maps a2 to all incoming bonds b
            # b2a maps each bond b to the atom it comes from a1
            # thus b2a[a2b] maps atom a2 to neighboring atoms a1
            self.a2a = self.b2a[self.a2b]  # num_atoms x max_num_bonds

        return self.a2a


def mol2graph(smiles_batch: List[str]) -> BatchMolGraph:
    """
    Converts a list of SMILES strings to a BatchMolGraph containing the batch of molecular graphs.
    :param smiles_batch: A list of SMILES strings.
    :return: A BatchMolGraph containing the combined molecular graph for the molecules
    """
    mol_graphs = []
    for smiles in smiles_batch:
        mol_graph = MolGraph(smiles)
        mol_graphs.append(mol_graph)

    return BatchMolGraph(mol_graphs)




import torch


def index_select_ND(source: torch.Tensor, index: torch.Tensor) -> torch.Tensor:
    """
    Selects the message features from source corresponding to the atom or bond indices in :code:`index`.
    :param source: A tensor of shape :code:`(num_bonds, hidden_size)` containing message features.
    :param index: A tensor of shape :code:`(num_atoms/num_bonds, max_num_bonds)` containing the atom or bond
                  indices to select from :code:`source`.
    :return: A tensor of shape :code:`(num_atoms/num_bonds, max_num_bonds, hidden_size)` containing the message
             features corresponding to the atoms/bonds specified in index.
    """
    index_size = index.size()  # (num_atoms/num_bonds, max_num_bonds)
    suffix_dim = source.size()[1:]  # (hidden_size,)
    final_size = index_size + suffix_dim  # (num_atoms/num_bonds, max_num_bonds, hidden_size)

    target = source.index_select(dim=0, index=index.view(-1))  # (num_atoms/num_bonds * max_num_bonds, hidden_size)
    target = target.view(final_size)  # (num_atoms/num_bonds, max_num_bonds, hidden_size)

    return target



import os
import setuptools

if os.path.exists('README.md'):
    long_description = open('README.md').read()
else:
    long_description = '''QM descriptor predictions through ChemProp'''

setuptools.setup(
    name="qmdesc", # Replace with your own username
    version="1.0.6",
    author="Yanfei Guan",
    author_email="yanfeig@mit.edu",
    description="A qm descriptor prediction package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    install_requires=[
        'torch>=1.4.0',
        'numpy>=1.18.1',
    ],
    url="https://github.com/yanfeiguan/qmdesc",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'': ['*.pt']},
    entry_points={
        'console_scripts': [
            'qmdesc=qmdesc:qmdesc',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    keywords=[
        'chemistry',
        'machine learning',
        'QM descriptors',
    ],

)



from argparse import Namespace

import torch.nn as nn

from qmdesc.mpn import MPN
from qmdesc.ffn import MultiReadout


class MoleculeModel(nn.Module):
    """A MoleculeModel is a model which contains a message passing network following by feed-forward layers."""

    def __init__(self, args, atom_fdim, bond_fdim):
        """
        Initializes the MoleculeModel.

        :param classification: Whether the model is a classification model.
        """
        super(MoleculeModel, self).__init__()
        self.create_encoder(args, atom_fdim, bond_fdim)
        self.create_ffn(args)

    def create_encoder(self, args: Namespace, atom_fdim, bond_fdim):
        """
        Creates the message passing encoder for the model.

        :param args: Arguments.
        """
        self.encoder = MPN(args, atom_fdim, bond_fdim)

    def create_ffn(self, args: Namespace):
        """
        Creates the feed-forward network for the model.

        :param args: Arguments.
        """

        # Create readout layer
        self.readout = MultiReadout(args, args.atom_targets, args.bond_targets,
                                    args.atom_constraints, args.bond_constraints)

    def forward(self, input):
        """
        Runs the MoleculeModel on input.

        :param input: Input.
        :return: The output of the MoleculeModel.
        """
        output_all = self.readout(self.encoder(input))

        return output_all
